"""
--- Day 23: Experimental Emergency Teleportation ---
https://adventofcode.com/2018/day/23
"""
from itertools import product

import pulp
from aocd import data
from parse import parse


pvar = pulp.LpVariable
template = "pos=<{:d},{:d},{:d}>, r={:d}"
bots = [parse(template, s).fixed for s in data.splitlines()]
prob = pulp.LpProblem("problem", pulp.LpMaximize)
cs = [pvar(f"c{i}", lowBound=0, upBound=1, cat="Integer") for i in range(len(bots))]
x = pvar("x")
y = pvar("y")
z = pvar("z")
n = pvar("n")

prob += n
prob += n == sum(cs)
for i, (xi, yi, zi, r) in enumerate(bots):
    ci = cs[i]
    for s in product([-1, 1], repeat=3):
        d = s[0] * (x - xi) + s[1] * (y - yi) + s[2] * (z - zi)
        prob += d <= r + (1 - ci) * 10 ** 10

status = prob.solve()
assert pulp.LpStatus[status] == "Optimal"
v = [int(pulp.value(c)) for c in (x, y, z)]
print("best location:", v)
print("bots in range:", int(pulp.value(n)))
print("distance from origin:", sum([abs(c) for c in v]))



"""
--- Day 4: Camp Cleanup ---
https://adventofcode.com/2022/day/4
"""
from aocd import data
from parse import parse

a = b = 0
for line in data.splitlines():
    l1, r1, l2, r2 = parse("{:d}-{:d},{:d}-{:d}", line).fixed
    if l1 > l2 or (l1 == l2 and r1 < r2):
        l1, r1, l2, r2 = l2, r2, l1, r1
    a += r2 <= r1
    b += r1 >= l2

print("part a:", a)
print("part b:", b)



"""
--- Day 1: Calorie Counting ---
https://adventofcode.com/2022/day/1
"""
from aocd import data
from heapq import nlargest

cals = []
for chunk in data.split("\n\n"):
    cals.append(sum(map(int, chunk.split())))

print("part a:", max(cals))
print("part b:", sum(nlargest(3, cals)))



"""
--- Day 3: Toboggan Trajectory ---
https://adventofcode.com/2020/day/3
"""
from math import prod
from aocd import data
from aoc_wim.zgrid import ZGrid

grid = ZGrid(data)
w, h = grid.width, grid.height
dzs = {}.fromkeys([1 + 1j, 3 + 1j, 5 + 1j, 7 + 1j, 1 + 2j], 0)
for dz in dzs:
    z = grid.top_left
    while z.imag < h:
        dzs[dz] += grid[complex(z.real % w, z.imag)] == "#"
        z += dz

print("part a:", dzs[3 + 1j])
print("part b:", prod(dzs.values()))



"""
--- Day 15: Dueling Generators ---
https://adventofcode.com/2017/day/15
"""
import numba
from aocd import data
from parse import parse


@numba.jit
def duel(a, b):
    fa = 16807
    fb = 48271
    d = 0x7fffffff
    count1 = count2 = 0
    a4 = []
    b8 = []
    for i in range(40000000):
        a *= fa
        a = (a & d) + (a >> 31)  # lo bits + hi bits
        a = (a & d) + (a >> 31)
        b *= fb
        b = (b & d) + (b >> 31)
        b = (b & d) + (b >> 31)
        a16 = a & 0xffff
        b16 = b & 0xffff
        count1 += a16 == b16
        if not a16 & 0b11:
            a4.append(a16)
        if not b16 & 0b111:
            b8.append(b16)
    while len(b8) < 5000000:
        b *= fb
        b = (b & d) + (b >> 31)
        b = (b & d) + (b >> 31)
        b16 = b & 0xffff
        if not b16 & 0b111:
            b8.append(b16)
    for i in range(5000000):
        count2 += a4[i] == b8[i]
    return count1, count2


template = "Generator A starts with {:d}\nGenerator B starts with {:d}"
a0, b0 = parse(template, data).fixed
a, b = duel(a0, b0)
print("part a:", a)
print("part b:", b)



"""Wim's solutions for https://adventofcode.com/"""
import logging
import runpy
import subprocess
import sys
from argparse import ArgumentParser
from datetime import datetime
from pathlib import Path

import aocd
from aocd.models import Puzzle
from aocd.utils import AOC_TZ


here = Path(__file__).parent


def run_one():
    parser = ArgumentParser(description=__doc__)
    aoc_now = datetime.now(tz=AOC_TZ)
    days = range(1, 26)
    years = range(2015, aoc_now.year + int(aoc_now.month == 12))
    parser.add_argument(
        "day",
        nargs="?",
        type=int,
        default=min(aoc_now.day, 25) if aoc_now.month == 12 else 1,
        help="1-25 (default: %(default)s)",
    )
    parser.add_argument(
        "year",
        nargs="?",
        type=int,
        default=years[-1],
        help="2015-%(default)s (default: %(default)s)",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-d", "--data", help="string or file to monkeypatch in aocd.data")
    group.add_argument("-t", "--test", help="use the example data (if any)", action="store_true")
    log_levels = "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"
    parser.add_argument("--log-level", type=str.upper, choices=log_levels)
    args = parser.parse_args()
    if args.day in years and args.year in days:
        args.day, args.year = args.year, args.day
    if args.log_level:
        level_int = getattr(logging, args.log_level)
        logging.basicConfig(format="%(message)s", level=level_int)
    mod_name = f"aoc_wim.aoc{args.year}.q{args.day:02d}"
    sys.modules.pop(mod_name, None)
    if args.data is not None:
        if args.data.endswith("txt"):
            path = Path(args.data)
            if path.is_file():
                txt = path.read_text()
                if "tests" in path.parts:
                    args.data = "\n".join(txt.splitlines()[:-2])
                else:
                    args.data = txt
        aocd.data = args.data
    elif args.test:
        test_dir = here.parent / "tests" / f"{args.year}" / f"{args.day:02d}"
        test_files = list(test_dir.glob("*.txt"))
        if not test_files:
            sys.exit(f"no test files found at {test_dir}")
        for test_path in sorted(test_files):
            print(f"\n(using data from {test_path})")
            print(f"--- {args.year} Day {args.day}: {Puzzle(args.year, args.day).title} ---")
            aocd.data = "\n".join(test_path.read_text().splitlines()[:-2])
            runpy.run_module(mod_name, run_name="__main__")
        print()
        return
    print(f"--- {args.year} Day {args.day}: {Puzzle(args.year, args.day).title} ---")
    runpy.run_module(mod_name, run_name="__main__")


def speedhack():
    here = Path(__file__).parent
    aoc_now = datetime.now(tz=AOC_TZ)
    if aoc_now.month != 12:
        sys.exit("It's not December yet")
    year = aoc_now.year
    day = aoc_now.day
    test_dir = here.parent / "tests" / str(year) / str(day).zfill(2)
    examples = sorted(test_dir.glob("*.txt"))
    puzzle = Puzzle(year, day)
    print(f"--- {year} Day {day}: {puzzle.title} ---")
    print(puzzle.url)
    if examples:
        print(f"{len(examples)} example(s) to test...")
        args = [sys.executable, "-m", "pytest", "-k", f"{year}/{day:02d}"]
        subprocess.check_call(args, cwd=Path("~/git/advent-of-code-wim").expanduser())
    args = ["aoc", "-y", str(year), "-d", str(day), "-u", "github.wimglenn.119932", "-r", "--log-level", "DEBUG", "--timeout", "600"]
    print("\n" + " ".join(args))
    subprocess.check_call(args)



"""
--- Day 3: Squares With Three Sides ---
https://adventofcode.com/2016/day/3
"""
import numpy as np
from aocd import data


A = np.fromstring(data, sep=" ", dtype=int)
R = A.reshape(-1, 3).T
C = R.reshape(-1, 3).T

for A in R, C:
    T = A[0] + A[1] > A[2]
    T &= A[0] + A[2] > A[1]
    T &= A[1] + A[2] > A[0]
    print(T.sum())



"""
--- Day 7: Some Assembly Required ---
https://adventofcode.com/2015/day/7
"""
import operator
import re

from aocd import data
import numpy as np


# TODO: change to use topological sort

opmap = {
    "AND": operator.and_,
    "OR": operator.or_,
    "LSHIFT": operator.lshift,
    "RSHIFT": operator.rshift,
}


def compute(data):
    result = {}

    def getval(v):
        return result[v] if v in result else np.uint16(v)

    lines = [line.split(" -> ") for line in data.splitlines()]

    def process_line(line):
        left, right = line
        left = left.split()
        len_left = len(left)
        if len_left == 1:
            # store
            result[right] = getval(left[0])
        elif len_left == 2:
            # negation
            op, val = left
            if op != "NOT":
                raise Exception
            result[right] = ~getval(val)
        elif len_left == 3:
            a, op, b = left
            op = opmap[op]
            result[right] = op(getval(a), getval(b))

    while lines:
        line = lines.pop()
        try:
            process_line(line)
        except (KeyError, ValueError):
            lines = [line] + lines

    return result


if __name__ == "__main__":
    result = compute(data)
    result_a = result["a"]
    print("part a:", result_a)

    new_data = re.sub(r"\n([0-9]+) -> b\n", "\n{} -> b\n".format(result_a), data)
    print("part b:", compute(new_data)["a"])



import logging
from collections import defaultdict
from collections import deque


# logging.basicConfig(level=logging.INFO, format="    %(message)s")
log = logging.getLogger(__name__)

# "parameter modes"
POSITION = 0
IMMEDIATE = 1
RELATIVE = 2


class IntComputer:
    class Halt(Exception):
        pass

    def __init__(self, reg0, inputs=()):
        self.ip = 0
        if reg0 and isinstance(reg0, str):
            reg0 = [int(x) for x in reg0.split(",")]
        self.offset = 0
        self.reg = defaultdict(int, dict(enumerate(reg0)))
        self.input = deque(inputs)
        self.output = deque()
        self._iterations = 0
        self._last_instruction = None
        self.op_map = {
            1: self.op_add,
            2: self.op_mul,
            3: self.op_input,
            4: self.op_output,
            5: self.op_jump_t,
            6: self.op_jump_f,
            7: self.op_lt,
            8: self.op_eq,
            9: self.op_offset,
            99: self.op_halt,
        }

    def op_add(self, x, y, z):
        self.reg[z] = self.reg[x] + self.reg[y]

    def op_mul(self, x, y, z):
        self.reg[z] = self.reg[x] * self.reg[y]

    def op_input(self, x):
        self.reg[x] = self.input.pop()

    def op_output(self, x):
        self.output.appendleft(self.reg[x])

    def op_jump_t(self, x, y):
        if self.reg[x]:
            self.ip = self.reg[y] - 3

    def op_jump_f(self, x, y):
        if not self.reg[x]:
            self.ip = self.reg[y] - 3

    def op_lt(self, x, y, z):
        self.reg[z] = int(self.reg[x] < self.reg[y])

    def op_eq(self, x, y, z):
        self.reg[z] = int(self.reg[x] == self.reg[y])

    def op_offset(self, x):
        self.offset += self.reg[x]

    def op_halt(self):
        raise IntComputer.Halt

    def step(self):
        opcode = self.reg[self.ip]
        op = self.op_map[opcode % 100]
        nargs = op.__func__.__code__.co_argcount
        args = []
        for i in range(1, nargs):
            mode = opcode // (10 ** (i + 1)) % 10
            if mode == IMMEDIATE:
                args.append(self.ip + i)
            elif mode == POSITION:
                args.append(self.reg[self.ip + i])
            elif mode == RELATIVE:
                args.append(self.reg[self.ip + i] + self.offset)
        e = "%4d %-10s ip=%-4d opcode=%05d  offset=%-5d args=%s"
        log.debug(e, self._iterations, op.__name__, self.ip, opcode, self.offset, args)
        self._last_instruction = op.__func__
        op(*args)
        self._iterations += 1
        self.ip += nargs

    @property
    def output_as_text(self):
        return "".join([chr(c) for c in reversed(self.output)])

    def input_text(self, value):
        for char in value:
            self.input.appendleft(ord(char))

    def run(self, until=None):
        if self.ip not in self.reg:
            return
        while True:
            try:
                self.step()
            except IntComputer.Halt:
                if until not in {None, IntComputer.Halt, IntComputer.op_halt}:
                    raise
                break
            if self._last_instruction is until:
                return

    def freeze(self):
        return self.ip, self.offset, tuple(self.reg.items())

    def unfreeze(self, state):
        self.ip, self.offset, reg = state
        self.reg.clear()
        for k, v in reg:
            self.reg[k] = v



import sys


def parsed(raw):
    parsed.n_bytes = len(raw)
    parsed.n_lines = len(raw.splitlines())
    this_module = sys.modules[__name__]
    for name, f in vars(this_module).items():
        if not name.startswith("_") and f is not parsed and getattr(f, "__module__", None) == this_module.__name__:
            try:
                result = f(raw)
            except Exception:
                pass
            else:
                parsed.parser = f.__doc__ or f.__name__
                return result
    parsed.parser = "raw"
    return raw


def space_separated_numbers(data):
    return [int(n) for n in data.split()]


def comma_separated_numbers(data):
    return [int(n) for n in data.split(",")]



"""
--- Day 14: Extended Polymerization ---
https://adventofcode.com/2021/day/14
"""
from collections import Counter
from aocd import data

polymer, rules = data.split("\n\n")
tr = dict(r.split(" -> ") for r in rules.splitlines())
c0 = Counter(x + y for x, y in zip(polymer, polymer[1:]))
c = Counter(polymer)
for i in range(40):
    if i == 10:
        print("part a:", max(c.values()) - min(c.values()))
    c1 = Counter()
    for xy, n in c0.items():
        c1[xy[0] + tr[xy]] += n
        c1[tr[xy] + xy[1]] += n
        c += {tr[xy]: n}
    c0 = c1
print("part b:", max(c.values()) - min(c.values()))



"""
--- Day 15: Oxygen System ---
https://adventofcode.com/2019/day/15
"""
from collections import deque
from aocd import data
from aoc_wim.aoc2019 import IntComputer
from aoc_wim.search import AStar
from aoc_wim.zgrid import ZGrid


NACK = 0
ACK = 1
GOAL = 2


neighbours = {
    -1j: 1,  # NORTH
    +1j: 2,  # SOUTH
    -1: 3,  # WEST
    +1: 4,  # EAST
}


class Q15AStar(AStar):
    def __init__(self, data):
        self.grid = ZGrid({0j: "."})
        self.comp = IntComputer(data)
        self.comp.output = deque(maxlen=1)
        self.freezer = {0j: self.comp.freeze()}
        state0 = 0j
        AStar.__init__(self, state0, None)

    def adjacent(self, z):
        self.comp.unfreeze(self.freezer[z])
        for dz, input_val in neighbours.items():
            self.comp.input.append(input_val)
            self.comp.run(until=IntComputer.op_output)
            [rc] = self.comp.output
            if rc == NACK:
                self.grid[z + dz] = "#"
            else:
                self.grid[z + dz] = "."
                self.freezer[z + dz] = self.comp.freeze()
                back = neighbours[-dz]
                self.comp.input.append(back)
                self.comp.run(until=IntComputer.op_output)
                assert self.comp.output[0]
                if rc == GOAL:
                    self.target = z + dz
                yield z + dz

    def draw(self):
        overlay = {0: "O"}
        if self.target is not None:
            overlay[self.target] = "T"
        self.grid.draw(overlay=overlay)


search = Q15AStar(data)
path = search.run()
print("part a", search.path_length)
search.draw()

s0 = search.freezer[search.target]
search.state0 = search.target
search.target_reached = lambda *args: False
search.comp.unfreeze(s0)
search.fscore.clear()
search.gscore.clear()
search.gscore[search.state0] = 0
search.came_from.clear()
search.closed.clear()
search.run()
print("part b", max(search.gscore.values()))
search.draw()



"""
--- Day 7: No Space Left On Device ---
https://adventofcode.com/2022/day/7
"""
from aocd import data
from collections import deque


q = deque(data.splitlines())
assert q.popleft() == "$ cd /"
cwd = "/"
dirs = {"/": 0}  # map of dirs to their cumulative size
parents = []  # cwd stack
while q:
    line = q.popleft()
    if line.startswith("$ cd "):
        _, _, name = line.split()
        if name == "..":
            cwd = parents.pop()
        else:
            parents.append(cwd)
            cwd += name + "/"
    elif line == "$ ls":
        while q and not q[0].startswith("$"):
            s, name = q.popleft().split()
            if s == "dir":
                dirs[cwd + name + "/"] = 0
            else:
                size = int(s)
                for d in parents + [cwd]:
                    dirs[d] += size

total_space = 70_000_000
df_target = 30_000_000
used_space = dirs["/"]
df_actual = total_space - used_space
del_needed = df_target - df_actual

a = sum(v for v in dirs.values() if v <= 100_000)
b = min(v for v in dirs.values() if v >= del_needed)

print("part a:", a)
print("part b:", b)



"""
--- Day 20: Infinite Elves and Infinite Houses ---
https://adventofcode.com/2015/day/20
"""
import numpy as np
from aocd import data


A = np.zeros(1000000, dtype=int)
B = A.copy()
for i in range(1, 1000000):
    A[i::i] += i * 10
    B[i::i][:50] += i * 11


assert A[1:10].tolist() == [10, 30, 40, 70, 60, 120, 80, 150, 130]
print("part a:", np.argmax(A > int(data)))
print("part b:", np.argmax(B > int(data)))



"""
--- Day 5: A Maze of Twisty Trampolines, All Alike ---
https://adventofcode.com/2017/day/5
"""
from aocd import data


def part_a(data, b_offset=1):
    state = [int(x) for x in data.split()]
    i = pos = 0
    while True:
        try:
            offset = state[pos]
        except IndexError:
            part_a.final_state = state
            return i
        i += 1
        state[pos] += b_offset if offset >= 3 else 1
        pos += offset


def part_b(data):
    return part_a(data, b_offset=-1)


print("part a:", part_a(data))
print("part b:", part_b(data))



"""
--- Day 6: Lanternfish ---
https://adventofcode.com/2021/day/6
"""
from aocd import data


def rotate(fish, n=1):
    for _ in range(n):
        fish[:] = fish[1:] + fish[:1]
        fish[6] += fish[-1]
    return sum(fish)


fish = [data.count(f) for f in "012345678"]
print("part a:", rotate(fish, n=80))
print("part b:", rotate(fish, n=256-80))



"""
--- Day 21: Allergen Assessment ---
https://adventofcode.com/2020/day/21
"""
from aocd import data
from bidict import bidict

d = {}
all_ingredients = []
for line in data.splitlines():
    ingredients, allergens = line.rstrip(")").split(" (contains ")
    ingredients = ingredients.split()
    all_ingredients += ingredients
    allergens = allergens.split(", ")
    for allergen in allergens:
        if allergen not in d:
            d[allergen] = set(ingredients)
        else:
            d[allergen] &= set(ingredients)

identified = bidict()
while d:
    for allergen, ingredients in d.items():
        if len(ingredients) == 1:
            break
    else:
        break
    [ingredient] = ingredients
    identified[allergen] = ingredient
    for ingredients in d.values():
        ingredients -= {ingredient}

print("part a:", sum(i not in identified.inv for i in all_ingredients))
print("part b:", ",".join(sorted(identified.inv, key=identified.inv.get)))



"""
--- Day 18: Duet ---
https://adventofcode.com/2017/day/18
"""
from aocd import data
from collections import deque
from aoc_wim.aoc2017 import Comp

# import logging; logging.basicConfig(format="%(message)s", level=logging.DEBUG)

comp = Comp(data, ch_rcv=deque([""]))
while comp.ch_rcv and not comp.blocked():
    comp.step()
print("part a:", comp.ch_snd[-1])

comp0 = Comp(data, pid=0)
comp1 = Comp(data, pid=1)
comp0.ch_snd = comp1.ch_rcv
comp1.ch_snd = comp0.ch_rcv
while not (comp0.blocked() and comp1.blocked()):
    if not comp0.blocked():
        comp0.step()
    if not comp1.blocked():
        comp1.step()
print("part b:", comp1.n_snd)



"""
--- Day 3: Perfectly Spherical Houses in a Vacuum ---
https://adventofcode.com/2015/day/3
"""
from aocd import data


step = {
    "^": -1j,
    ">": 1,
    "v": 1j,
    "<": -1,
}


z = 0
seen = {z}
for c in data:
    z += step[c]
    seen |= {z}

print("part a:", len(seen))


z = 0
seen = {z}
for c in data[0::2]:  # santa
    z += step[c]
    seen |= {z}

z = 0
for c in data[1::2]:  # robo-santa
    z += step[c]
    seen |= {z}

print("part b:", len(seen))



"""
--- Day 12: Leonardo's Monorail ---
https://adventofcode.com/2016/day/12
"""
from aocd import data
from aoc_wim.aoc2016 import AssembunnyComputer


comp = AssembunnyComputer(data)
comp.run()
print("part a:", comp.reg["a"])

comp.__init__(data, c0=1)
comp.run()
print("part b:", comp.reg["a"])



"""
--- Day 23: Crab Cups ---
https://adventofcode.com/2020/day/23
"""
from aocd import data
from operator import attrgetter


class Cup:

    __slots__ = "label", "r"

    def __init__(self, label, r=None):
        self.label = label
        self.r = r  # the cup clockwise to the right


class CupGame:

    def __init__(self, data, n=None):
        unsorted = len(data)
        labels = [int(x) for x in data.strip()]
        self.maxlabel = max(labels)
        if n is not None:
            labels.extend(range(self.maxlabel + 1, n + 1))
            self.maxlabel = labels[-1]
        cups = self.cups = [Cup(n) for n in labels]
        for i in range(self.maxlabel):
            cups[i - 1].r = cups[i]
        self.current = cups[0]
        self.cups[:unsorted] = sorted(self.cups[:unsorted], key=attrgetter("label"))

    def play(self, iterations=1):
        for i in range(iterations):
            label = self.current.label - 1 or self.maxlabel
            cut = self.current.r
            self.current.r = cut.r.r.r
            pickup = [
                cut.label,
                cut.r.label,
                cut.r.r.label,
            ]
            while label in pickup:
                label = label - 1 or self.maxlabel
            dest = self.cups[label - 1]
            cut.r.r.r, dest.r = dest.r, cut
            self.current = self.current.r


game = CupGame(data)
game.play(iterations=100)
cup1 = game.cups[0]
cup = cup1.r
labels = []
while cup is not cup1:
    labels.append(str(cup.label))
    cup = cup.r
print("part a:", "".join(labels))

game = CupGame(data, n=1000000)
game.play(iterations=10000000)
cup1 = game.cups[0]
print("part b:", cup1.r.label * cup1.r.r.label)



"""
--- Day 16: Packet Decoder ---
https://adventofcode.com/2021/day/16
"""
import io
import math
from aocd import data


def literal():
    pass


def gt(vals):
    a, b = vals
    return 1 if a > b else 0


def lt(vals):
    a, b = vals
    return 1 if a < b else 0


def eq(vals):
    a, b = vals
    return 1 if a == b else 0


ops = {
    "+": sum,
    "*": math.prod,
    "min": min,
    "max": max,
    "val": literal,
    ">": gt,
    "<": lt,
    "==": eq,
}
op_types = [*ops.items()]


class Packet:

    version_sum = 0

    def __init__(self, stream, indent=0):
        self.indent = indent
        self.n_read = 0
        self.stream = stream
        self.version = self.read_int(3)
        Packet.version_sum += self.version
        type_id = self.read_int(3)
        glyph, self.func = op_types[type_id]
        self.subpackets = []
        if self.func is literal:
            nibbles = []
            prefix = 1
            while prefix:
                prefix = self.read_int(1)
                nibbles.append(self.read_raw(4))
            self.value = int("".join(nibbles), 2)
        else:  # operator packet
            self.length_type_id = self.read_int(1)
            if self.length_type_id == 0:
                total_length = self.read_int(15)
                subpackets_raw = self.read_raw(total_length)
                substream = io.StringIO(subpackets_raw)
                while total_length:
                    try:
                        subpacket = Packet(substream, indent=self.indent + 1)
                    except EOFError:
                        break
                    total_length -= subpacket.n_read
                    self.subpackets.append(subpacket)
            else:
                number_of_subpackets = self.read_int(11)
                while number_of_subpackets:
                    try:
                        subpacket = Packet(self.stream, indent=self.indent + 1)
                    except EOFError:
                        break
                    number_of_subpackets -= 1
                    self.subpackets.append(subpacket)
            vals = [p.value for p in self.subpackets]
            self.value = self.func(vals)

    def read_raw(self, nbits):
        raw = self.stream.read(nbits)
        if len(raw) < nbits:
            raise EOFError
        self.n_read += nbits
        return raw

    def read_int(self, nbits):
        return int(self.read_raw(nbits), 2)

    def __str__(self):
        return "\t" * self.indent + f"<Packet v={self.version}, op={self.func.__name__}, val={self.value}>"


stream = io.StringIO("".join([f"{int(x, 16):04b}" for x in data]))
packet0 = Packet(stream)
print("part a:", Packet.version_sum)
print("part b:", packet0.value)



"""
--- Day 22: Mode Maze ---
https://adventofcode.com/2018/day/22
"""
from aocd import data
from parse import parse
from aoc_wim.search import AStar
from aoc_wim.zgrid import zrange
from aoc_wim.zgrid import manhattan_distance


# TODO: use zgrid


class Grid:

    glyph = dict(enumerate(".=|"))  # rocky, wet, narrow

    def __init__(self, depth, tx, ty):
        self.depth = depth
        self.pos = 0
        self.target = complex(tx, ty)
        self._gi = {self.pos: 0, self.target: 0}

    def el(self, pos):
        return (self.gi(pos) + self.depth) % 20183

    def t(self, pos):
        return self.el(pos) % 3

    def gi(self, pos):
        if pos not in self._gi:
            if pos.imag == 0:
                self._gi[pos] = 16807 * int(pos.real)
            elif pos.real == 0:
                self._gi[pos] = 48271 * int(pos.imag)
            else:
                self._gi[pos] = self.el(pos - 1) * self.el(pos - 1j)
        return self._gi[pos]


def parsed(data):
    template = "depth: {:d}\ntarget: {:d},{:d}"
    d, x, y = parse(template, data)
    grid = Grid(d, x, y)
    return grid


def draw(grid, z0=0, z1=None):
    if z1 is None:
        z1 = grid.target + 6 + 6j
    for z in zrange(z0, z1):
        if z == grid.pos:
            g = "M"
        elif z == grid.target:
            g = "T"
        else:
            g = Grid.glyph[grid.t(z)]
        print(g, end="")
        if z.real + 1 == z1.real:
            print()


class Q22AStar(AStar):

    tools = {None, "🔦", "🧗"}

    valid_tools = {
        ".": tools - {None},  # you'll likely slip and fall
        "=": tools - {"🔦"},  # if it gets wet, you won't have a light source
        "|": tools - {"🧗"},  # it's too bulky to fit
    }

    def __init__(self, state0, target, grid):
        super().__init__(state0, target)
        self.grid = grid

    def heuristic(self, state0, state1):
        z0, tool0 = state0
        z1, tool1 = state1
        result = manhattan_distance(z1, z0)
        if tool0 != tool1:
            result += 7
        return result

    def cost(self, current_state, next_state):
        z0, tool0 = current_state
        z1, tool1 = next_state
        if z0 == z1:
            assert tool0 != tool1
            return 7
        else:
            assert abs(z1 - z0) == 1
            return 1

    def adjacent(self, state):
        # 012   .=|   rocky, wet, narrow   no neither, no torch, no climbing gear
        # None, "🔦", "🧗"
        z_here, tool = state
        current_terain = Grid.glyph[self.grid.t(z_here)]
        assert tool in self.valid_tools[current_terain]
        zs = [z_here + 1, z_here + 1j]
        if z_here.real > 0:
            zs.append(z_here - 1)
        if z_here.imag > 0:
            zs.append(z_here - 1j)
        terrains = [Grid.glyph[self.grid.t(z)] for z in zs]
        zs = [z for z, t in zip(zs, terrains) if tool in self.valid_tools[t]]
        zs = [(z, tool) for z in zs]
        tools = Q22AStar.tools - {tool}
        zs.extend([(z_here, t) for t in tools if t in self.valid_tools[current_terain]])
        return zs


grid = parsed(data)
print("part a:", sum(grid.t(z) for z in zrange(0, grid.target + 1 + 1j)))

state0 = (grid.pos, "🔦")
target = (grid.target, "🔦")
a_star = Q22AStar(state0=state0, target=target, grid=grid)
a_star.run()
print("part b:", a_star.gscore[target])



"""
--- Day 25: The Halting Problem ---
https://adventofcode.com/2017/day/25
"""
from collections import defaultdict

from aocd import data
from parse import parse


template_first = """\
Begin in state {state0}.
Perform a diagnostic checksum after {n:d} steps."""


template_rest = """\
In state {state_current}:
  If the current value is 0:
    - Write the value {val0:d}.
    - Move one slot to the {direction0}.
    - Continue with state {state0}.
  If the current value is 1:
    - Write the value {val1:d}.
    - Move one slot to the {direction1}.
    - Continue with state {state1}."""


first, *rest = data.split("\n\n")
prog = parse(template_first, first).named
directions = {"right": 1, "left": -1}
for text in rest:
    data = parse(template_rest, text).named
    prog[data["state_current"]] = {
        0: (directions[data["direction0"]], data["val0"], data["state0"]),
        1: (directions[data["direction1"]], data["val1"], data["state1"]),
    }

tape = defaultdict(int)
cursor = 0
state = prog["state0"]
for i in range(prog["n"]):
    instructions = prog[state]
    val = tape[cursor]
    direction, tape[cursor], state = instructions[val]
    cursor += direction

print(sum(tape.values()))



"""
--- Day 17: Two Steps Forward ---
https://adventofcode.com/2016/day/17
"""
from collections import deque
from _md5 import md5
from aocd import data


def bfs(part="a"):
    z0 = 0
    target = 3 + 3j
    depth = 0
    path = ""
    longest_path_length = None
    queue = deque([(z0, path, depth)])
    while queue:
        z, path, depth = queue.popleft()
        if z == target:
            if part == "a":
                return path
            longest_path_length = depth
        else:
            queue.extend((x + (depth + 1,)) for x in adjacent(z, path))
    return longest_path_length


def adjacent(z0, path):
    dzs = {"U": -1j, "D": 1j, "L": -1, "R": 1}
    directions = zip(dzs.items(), md5((data + path).encode()).hexdigest())
    for (dpath, dz), s in directions:
        if s in "bcdef":
            z = z0 + dz
            if 0 <= z.real <= 3 and 0 <= z.imag <= 3:
                yield z, path + dpath


print("part a:", bfs(part="a"))
print("part b:", bfs(part="b"))



"""
--- Day 13: A Maze of Twisty Little Cubicles ---
https://adventofcode.com/2016/day/13
"""
from aoc_wim.zgrid import ZGrid
from aocd import data


class WallMap:
    def __init__(self, fav_number=None):
        if fav_number is None:
            fav_number = int(data)
        self.fav_number = fav_number

    def __call__(self, z):
        if z.real < 0 or z.imag < 0:
            return "#"
        x, y = z.real, z.imag
        fz = x * x + 3 * x + 2 * x * y + y + y * y
        popcount = bin(int(fz) + self.fav_number).count("1")
        result = "#" if popcount % 2 else "."
        return result


z0 = 1 + 1j
target = 31 + 39j


if __name__ == "__main__":
    grid = ZGrid(WallMap(), on=".", off="#")
    depths = grid.bfs(target=target, z0=z0)
    print("part a:", depths[target])
    depths = grid.bfs(z0=z0, max_depth=50)
    print("part b:", len(depths))
    grid.draw_path(z=target, z0=z0)



"""
--- Day 14: Space Stoichiometry ---
https://adventofcode.com/2019/day/14
"""
from aocd import data
from collections import Counter
from aoc_wim.search import Bisect
import networkx as nx


def parsed(data):
    digraph = nx.DiGraph()
    for line in data.splitlines():
        sources, dest = line.split(" => ")
        n0, elem0 = dest.split()
        for source in sources.split(", "):
            n1, elem1 = source.split()
            digraph.add_edge(elem0, elem1, ratio=(int(n0), int(n1)))
    return digraph


def part_a(data, fuel=1):
    elems = Counter({"FUEL": fuel})
    digraph = parsed(data)
    for elem0 in nx.topological_sort(digraph):
        amount = elems.pop(elem0)
        if elem0 == "ORE":
            assert not elems
            return amount
        for elem1, edge_data in digraph[elem0].items():
            n0, n1 = edge_data["ratio"]
            r = -(amount // -n0)  # ceiling division
            elems[elem1] += r * n1


def part_b(data):
    n_ore = 1000000000000
    bisect = Bisect(lambda fuel: part_a(data, fuel), val=n_ore, lo=1)
    return bisect.run()


print("part a:", part_a(data))
print("part b:", part_b(data))



import heapq
from collections import defaultdict
from itertools import count
import logging


log = logging.getLogger(__name__)


class AStar:
    def __init__(self, state0, target):
        self.state0 = state0
        self.target = target
        self.closed = set()
        self.came_from = {}
        inf = float("inf")
        # optimistic estimate of cost to reach node
        self.fscore = defaultdict(lambda: inf, {target: self.heuristic(state0, target)})
        # shortest cost to reach node found so far
        self.gscore = defaultdict(lambda: inf, {state0: 0})
        self.path_length = None

    def heuristic(self, state0, state1):
        """estimate of distance (cost) to get from state0 to state1
        note: in order for the heuristic to be admissible, this must
        not overestimate the actual distance (i.e. be optimistic)
        """
        return 1

    def cost(self, current_state, next_state):
        """actual delta to get from current state to an adjacent state"""
        return 1

    def adjacent(self, state):
        """other states reachable from given state"""
        return []

    def reconstruct_path(self, current=None):
        if current is None:
            current = self.target
        path = [current]
        while current in self.came_from:
            current = self.came_from[current]
            path.append(current)
        return path

    def target_reached(self, current_state, target):
        return current_state == target

    def run(self):
        i = count()  # tie-breaker
        heap = [(0, next(i), self.state0)]
        while heap:
            _score, _id, current_state = heapq.heappop(heap)
            if self.target_reached(current_state, self.target):
                path_to_target = self.reconstruct_path(current_state)
                self.path_length = len(path_to_target) - 1
                if self.target is None:
                    self.target = current_state
                return path_to_target
            self.closed.add(current_state)
            for next_state in self.adjacent(current_state):
                if next_state in self.closed:
                    continue
                tentative_gscore = self.gscore[current_state]
                tentative_gscore += self.cost(current_state, next_state)
                if tentative_gscore >= self.gscore[next_state]:
                    continue
                self.came_from[next_state] = current_state
                self.gscore[next_state] = tentative_gscore
                fscore = tentative_gscore + self.heuristic(next_state, self.target)
                self.fscore[next_state] = fscore
                heapq.heappush(heap, (fscore, next(i), next_state))


class BisectError(Exception):
    pass


class Bisect:

    def __init__(self, callable, val=0.5, lo=None, hi=None):
        self.callable = callable
        self.val = val
        self.lo = lo
        self.hi = hi
        self.results = {}

    def discover_bounds(self):
        step = 1
        if self.lo is None:
            pos = 0
            if self.hi is not None:
                pos = self.hi - 1
            while True:
                self.results[pos] = result = self.callable(pos)
                log.debug("discovering lower bound %s: %s", pos, result)
                if result < self.val:
                    self.lo = pos
                    break
                if self.hi is not None:
                    self.hi = min(self.hi, pos)
                pos -= step
                step *= 2
        step = 1
        if self.hi is None:
            pos = self.lo + 1
            while True:
                self.results[pos] = result = self.callable(pos)
                log.debug("discovering upper bound %s: %s", pos, result)
                if self.val <= result:
                    self.hi = pos
                    break
                self.lo = max(self.lo, pos)
                pos += step
                step *= 2
        if self.lo > self.hi:
            raise BisectError("error discovering bounds")
        if self.results[self.lo] > self.results[self.hi]:
            raise BisectError("callable isn't non-decreasing between %s and %s", self.lo. self.hi)

    def step(self):
        mid = (self.lo + self.hi) // 2
        self.results[mid] = result = self.callable(mid)
        log.debug("bisected %s: %s", mid, result)
        if result < self.val:
            self.lo = mid
        elif self.val <= result:
            self.hi = mid

    def run(self):
        if self.lo is None or self.hi is None:
            self.discover_bounds()
        while self.hi - self.lo > 1:
            self.step()
        return self.lo



"""
--- Day 25: Let It Snow ---
https://adventofcode.com/2015/day/25
"""
from aocd import data


words = data.split()
row = int(words[-3].rstrip(","))
col = int(words[-1].rstrip("."))


def n(row, col):
    i = (row + col) * (row + col) + 2 - col - 3 * row
    return i // 2


m = 252533
d = 33554393
i = n(row, col)
code = 20151125
for _ in range(1, i):
    code = m * code % d

print(code)



"""
--- Day 4: Giant Squid ---
https://adventofcode.com/2021/day/4
"""
from aocd import data

draw, *boards = data.split("\n\n")
draw = [int(n) for n in draw.split(",")]
rows = [[[int(n) for n in ns.split()] for ns in b.splitlines()] for b in boards]
cols = [[*zip(*r)] for r in rows]
boards = [[set(r) for r in rs] + [set(c) for c in cs] for rs, cs in zip(rows, cols)]


def winner_score(draw, boards):
    # removes the winner's board from the game, and return their score
    for n in draw:
        for board in boards:
            for line in board:
                line.discard(n)
                if not line:  # this board got bingo
                    remaining_numbers = line.union(*board) - {n}
                    boards[:] = [b for b in boards if b is not board]
                    return sum(remaining_numbers) * n


print("part a:", winner_score(draw, boards))
while boards:
    b = winner_score(draw, boards)
print("part b:", b)



"""
--- Day 23: Experimental Emergency Teleportation ---
https://adventofcode.com/2018/day/23
"""
from heapq import heappop
from heapq import heappush
from itertools import product

import math
import numpy as np
from aocd import data
from parse import parse


def parsed(data):
    template = "pos=<{:d},{:d},{:d}>, r={:d}"
    nums = [parse(template, s).fixed for s in data.splitlines()]
    A = np.array(nums)
    xs = A[:, :-1]
    rs = A[:, -1]
    return xs, rs


class Cell:
    def __init__(self, x, s):
        self.x = x  # 3d position of corner
        self.s = s  # edge size
        lo = np.maximum(self.x - xs, 0)
        hi = np.maximum(xs - self.x - self.s, 0)
        self.n_out = ((lo + hi).sum(axis=1) > rs).sum()  # count bots out of range
        self.d = abs(self.x).sum()  # manhattan distance of corner from origin

    def split(self):
        # divide this 3D space into 8 evenly sized subspaces
        # see https://en.wikipedia.org/wiki/Octree for inspiration
        halfwidth = self.s // 2
        dx = np.array(list(product([0, 1], repeat=3))) * halfwidth
        for x in self.x + dx:
            yield Cell(x, halfwidth)

    def __lt__(self, other):
        # key for heapq comparison - we will minimise over tuples of:
        # (number of bots outside, cell edge size, distance of the corner from origin)
        # maximize number in range = minimizing number out of range
        return (self.n_out, self.s, self.d) < (other.n_out, other.s, other.d)


xs, rs = parsed(data)
i = rs.argmax()
print("part a:", (abs(xs - xs[i]).sum(axis=1) <= rs[i]).sum())

x0 = 2 ** math.ceil(math.log2(abs(xs).max()))
cell = Cell(x=np.array([-x0, -x0, -x0]), s=x0 * 2)
assert cell.n_out == 0, "initial cell must cover all bots"

pq = [cell]
while pq:
    cell = heappop(pq)
    print(f"qlen={len(pq)} s={cell.s} {len(xs) - cell.n_out} nanobots")
    if not cell.s:
        break
    for subcell in cell.split():
        heappush(pq, subcell)
print("part b:", cell.d)



"""
--- Day 15: Science for Hungry People ---
https://adventofcode.com/2015/day/15
"""
from aocd import data
from parse import parse


def parsed(data):
    template = "{name}: capacity {:d}, durability {:d}, flavor {:d}, texture {:d}, calories {:d}"
    ingredients = []
    for line in data.splitlines():
        nums = parse(template, line).fixed
        ingredients.append(nums)
    return ingredients


def divisions(n, target=100):
    # tuples of length n which sum to target
    if n == 1:
        yield target,
    else:
        for i in range(target + 1):
            yield from [(i,) + t for t in divisions(n - 1, target - i)]


def score(division, ingredients, cal_target=None):
    parts = [0] * len(ingredients[0])
    for div, ingredient in zip(division, ingredients):
        part = [n * div for n in ingredient]
        parts = [a + b for a, b in zip(parts, part)]
    result = 1
    for part in parts[:-1]:
        result *= max(0, part)
    if cal_target is not None and parts[-1] != cal_target:
        result = 0
    return result


def best_score(data, cal_target=None):
    best = 0
    ingredients = parsed(data)
    for div in divisions(len(ingredients)):
        best = max(best, score(div, ingredients, cal_target=cal_target))
    return best


print(best_score(data))
print(best_score(data, cal_target=500))



"""
--- Day 1: Chronal Calibration ---
https://adventofcode.com/2018/day/1
"""
from itertools import cycle

from aocd import data


ns = [int(x) for x in data.split()]
print("part a:", sum(ns))

f = None
if len({n > 0 for n in ns}) == 2:
    f = 0
    seen = {f}
    for n in cycle(ns):
        f += n
        if f in seen:
            break
        seen.add(f)
print("part b:", f)



"""
--- Day 13: Care Package ---
https://adventofcode.com/2019/day/13
"""
from collections import deque
from aocd import data
from aoc_wim.aoc2019 import IntComputer


class Game:
    def __init__(self, data):
        self.cpu = IntComputer(data)
        self.cpu.input = self
        self.cpu.output = deque(maxlen=3)
        self.paddle_position = 0
        self.ball_position = 0
        self.n_blocks = 0
        self.score = 0

    def pop(self):
        p = self.paddle_position
        b = self.ball_position
        return -1 if b < p else 0 if b == p else 1

    def step(self):
        self.cpu.run(until=IntComputer.op_output)
        self.cpu.run(until=IntComputer.op_output)
        self.cpu.run(until=IntComputer.op_output)
        val, y, x = self.cpu.output
        if val == 2:
            self.n_blocks += 1
        elif val == 3:
            self.paddle_position = x
        elif val == 4:
            self.ball_position = x
        if x == -1 and y == 0:
            self.score = val

    def play(self):
        while True:
            try:
                self.step()
            except IntComputer.Halt:
                return


game_a = Game(data)
game_a.play()
print("part a:", game_a.n_blocks)

game_b = Game(data)
game_b.cpu.reg[0] = 2
game_b.play()
print("part b:", game_b.score)
